(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir

   $Id: EqDecInstances.v 2768 2009-02-27 01:35:59Z baydemir $ *)

(** A library that declares hints, facts, and instances for [EqDec]
    for some common types. *)

Require Import Coq.Arith.Peano_dec.
Require Import Coq.Classes.EquivDec.
Require Import Coq.Logic.Decidable.
Require Import Coq.Strings.String.


(* *********************************************************************** *)
(** * Facts *)

Lemma equiv_decidable : forall `{EqDec A} (x y : A),
  decidable (x === y).
Proof.
  unfold decidable.
  intros ? ? ? ? x y.
  destruct (x == y); auto.
Defined.


(* *********************************************************************** *)
(** * Tactic support *)

(* The following tactic should be able to prove an [EqDec _ eq]
   instance directly. *)

Ltac EqDec_eq_tac :=
  compute;
  let x := fresh in
  let y := fresh in
  intros x y;
  fold (~ (x = y));
  repeat (auto; intros; decide equality);
  auto.

(** It is useful for the [auto] tactic to invoke [reflexivity] and
    [symmetry] when working with a setoid equality. *)

Hint Extern 1 (_ === _) => reflexivity.
Hint Extern 1 (_ === _) => symmetry.
Hint Extern 1 (_ =/= _) => symmetry.

(** [nat] and [string] both have readily available proofs of decidable
    equality. *)

Hint Resolve eq_nat_dec.
Hint Resolve string_dec.


(* *********************************************************************** *)
(** * Instances *)

(** [nat] and [string] are instances of [EqDec]. *)

Instance EqDec_nat_eq : EqDec nat eq.
(* begin show *)
Proof. EqDec_eq_tac. Defined.
(* end show *)

Instance EqDec_string_eq : EqDec string eq.
(* begin show *)
Proof. EqDec_eq_tac. Defined.
(* end show *)
